#version 330
#extension GL_ARB_separate_shader_objects : enable





layout(location = 0) smooth in vec2 _fin_fragTexCoord;

struct SFragmentInput
{
	vec2 fragTexCoord;
	bool IsFrontFace;
};

layout(location = 0) out vec4 _fout_outColor;

struct SFragmentOutput
{
	vec4 outColor;
};

 uniform sampler2D Texture;

layout( std140) uniform UPushConstantBuffer
{
	ivec2 Resolution;
} PushConstantBuffer;

vec3 FXAA( vec2 texCoords,
					float subpix,
					float edgeThreshold,
					float edgeThresholdMin
					)
{
	vec2 resolution = vec2( PushConstantBuffer.Resolution );
	vec2 inverseResolution = 1.0f / resolution;
	float FXAAQualityP[] =  float[] (  1.0, 1.0, 1.0, 1.0, 1.0, 1.5, 2.0, 2.0, 2.0, 2.0, 4.0, 8.0  ) ;






	vec3 rgbM =  texture(Texture, texCoords) .rgb;
	float lumaM = dot(pow(rgbM, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114));


	float lumaN = dot(pow( texture(Texture, texCoords + vec2(0.0, +1.0) * inverseResolution) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) ;
	float lumaS = dot(pow( texture(Texture, texCoords + vec2(0.0, -1.0) * inverseResolution) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) ;
	float lumaW = dot(pow( texture(Texture, texCoords + vec2(-1.0, 0.0) * inverseResolution) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) ;
	float lumaE = dot(pow( texture(Texture, texCoords + vec2(+1.0, 0.0) * inverseResolution) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) ;

	float rangeMin = min(lumaM, min(min(lumaN, lumaS), min(lumaE, lumaW)));
	float rangeMax = max(lumaM, max(max(lumaN, lumaS), max(lumaE, lumaW)));

	float range = rangeMax - rangeMin;

	if (range < max(edgeThresholdMin, rangeMax * edgeThreshold))
		return rgbM;

	float lumaNE = dot(pow( texture(Texture, texCoords + vec2(+1.0, +1.0) * inverseResolution) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) ;
	float lumaNW = dot(pow( texture(Texture, texCoords + vec2(-1.0, +1.0) * inverseResolution) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) ;
	float lumaSW = dot(pow( texture(Texture, texCoords + vec2(-1.0, -1.0) * inverseResolution) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) ;
	float lumaSE = dot(pow( texture(Texture, texCoords + vec2(+1.0, -1.0) * inverseResolution) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) ;

	float lumaNS = lumaN + lumaS;
	float lumaWE = lumaW + lumaE;
	float lumaNESE = lumaNE + lumaSE;
	float lumaNWNE = lumaNW + lumaNE;
	float lumaNWSW = lumaNW + lumaSW;
	float lumaSWSE = lumaSW + lumaSE;

	float edgeHorz = abs(-2.0 * lumaW + lumaNWSW) + abs(-2.0 * lumaM + lumaNS) * 2.0 + abs(-2.0 * lumaE + lumaNESE);
	float edgeVert = abs(-2.0 * lumaN + lumaNWNE) + abs(-2.0 * lumaM + lumaWE) * 2.0 + abs(-2.0 * lumaS + lumaSWSE);

	bool horzSpan = (edgeHorz >= edgeVert);

	float luma0 = horzSpan ? lumaS : lumaW;
	float luma1 = horzSpan ? lumaN : lumaE;

	float gradient0 = luma0 - lumaM;
	float gradient1 = luma1 - lumaM;

	float gradientScaled = 0.25f * max(abs(gradient0), abs(gradient1));

	float lengthSign = horzSpan ? inverseResolution.y : inverseResolution.x;

	float lumaA = 0.0f;
	if (abs(gradient0) >= abs(gradient1))
	{
		lengthSign = -lengthSign;
		lumaA = 0.5f * (luma0 + lumaM);
	}
	else
	{
		lumaA = 0.5f * (luma1 + lumaM);
	}

	vec2 posB = texCoords;
	if (horzSpan)
		posB.y += lengthSign * 0.5f;
	else
		posB.x += lengthSign * 0.5f;

	vec2 offset = horzSpan ? vec2(inverseResolution.x, 0.0f) : vec2(0.0f, inverseResolution.y);

	vec2 pos0 = posB - offset * FXAAQualityP[0];
	vec2 pos1 = posB + offset * FXAAQualityP[0];

	float lumaEnd0 = dot(pow( texture(Texture, pos0) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) - lumaA;
	float lumaEnd1 = dot(pow( texture(Texture, pos1) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) - lumaA;

	bool done0 = abs(lumaEnd0) >= gradientScaled;
	bool done1 = abs(lumaEnd1) >= gradientScaled;
	bool done01 = done0 && done1;

	if (!done0)
		pos0 -= offset * FXAAQualityP[1];
	if (!done1)
		pos1 += offset * FXAAQualityP[1];

	for (int i = 2; (i < 12) && !done01; i++)
	{
		if (!done0)
		{
			lumaEnd0 = dot(pow( texture(Texture, pos0) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) - lumaA;
		}
		if (!done1)
		{
			lumaEnd1 = dot(pow( texture(Texture, pos1) .rgb, vec3(1.0f, 1.0f, 1.0f) / 2.2f), vec3(0.299, 0.587, 0.114)) - lumaA;
		}
		done0 = abs(lumaEnd0) >= gradientScaled;
		done1 = abs(lumaEnd1) >= gradientScaled;
		done01 = done0 && done1;

		if(!done0)
			pos0 -= offset * FXAAQualityP[i];
		if(!done1)
			pos1 += offset * FXAAQualityP[i];

		if (done01)
			break;
	}

	float dist0 = horzSpan ? (texCoords.x - pos0.x) : (texCoords.y - pos0.y);
	float dist1 = horzSpan ? (pos1.x - texCoords.x) : (pos1.y - texCoords.y);

	bool isDirection0 = dist0 < dist1;
	float dist = min(dist0, dist1);

	float spanLength = dist0 + dist1;

	float pixelOffset = -(dist / spanLength) + 0.5;

	bool lumaMLTZero = lumaM < lumaA;

	bool goodSpan = ((isDirection0 ? lumaEnd0 : lumaEnd1) < 0.0) != lumaMLTZero;
	float finalOffset = goodSpan ? pixelOffset : 0.0f;

	float lumaAverage = (1.0f / 12.0f) * (2.0f * (lumaNS + lumaWE) + lumaNWSW + lumaNESE);

	float subPixelOffset0 = clamp(abs(lumaAverage - lumaM) / range, 0.0f, 1.0f);
	float subPixelOffset1 = (-2.0f * subPixelOffset0 + 3.0f) * subPixelOffset0 * subPixelOffset0;
	float subPixelOffsetFinal = subPixelOffset1 * subPixelOffset1 * subpix;

	finalOffset = max(finalOffset, subPixelOffsetFinal);

	if (horzSpan)
		texCoords.y += finalOffset * lengthSign;
	else
		texCoords.x += finalOffset * lengthSign;

	return  texture(Texture, texCoords) .rgb;
}

void FragmentMain( in SFragmentInput  fInput,  out SFragmentOutput  fOutput  )
{
	vec2 texCoords = fInput.fragTexCoord;
	vec3 preset = vec3(0.75f, 0.125f, 0.0625f);
	fOutput.outColor.rgb = FXAA(texCoords, preset.x, preset.y, preset.z  );
	fOutput.outColor.a = 1.0f;
}


void main()
{
	SFragmentInput fInput;
	SFragmentOutput fOutput;

	fInput.fragTexCoord = _fin_fragTexCoord;
	fInput.IsFrontFace = gl_FrontFacing;
	FragmentMain(fInput, fOutput);
	_fout_outColor = fOutput.outColor;
}
